module BugfixOutputToStdErr.Fib where

-- | Calculate Fibonacci number of given 'Num'.
--
-- >>> import System.IO
-- >>> hPutStrLn stderr "foobar"
-- foobar
fib :: (Num t, Num t1) => t -> t1
fib _ = undefined
